;---------------------------------------------------------------------------
;          TILD-ESC.ASM
;          Memory resident; Swaps role of <Tilde> and <Escape> keys.
;          Approach is that this routine will handle all processing
;             associated with the <Escape> and <Tilde> keys, and
;             that the BIOS routine will handle all other keys.
;          Must be converted to a .COM file.
;---------------------------------------------------------------------------
;
;------ Equates
KBDATA     EQU       60H               ; Keyboard data port
KBCTRL     EQU       61H               ; Keyboard control port
EOI        EQU       20H               ; 8259 EOI value
INT_CTRL   EQU       20H               ; 8259 port address
;
;======================================= Segment BIOS_DATA
;
BIOS_DATA  SEGMENT   AT 40H
           ORG       17H
KB_STATUS  DB        ?                 ; Keyboard status byte
           ORG       1AH
BUF_HEAD   DW        ?                 ; Pointer to keyboard buffer head
BUF_TAIL   DW        ?                 ; Pointer to keyboard buffer tail
           ORG       80H
BUF_START  DW        ?                 ; Start keyboard buffer address
BUF_END    DW        ?                 ; End keyboard buffer address plus two
BIOS_DATA  ENDS
;
;======================================= Segment COMSEG
;
COMSEG     SEGMENT   PARA 'CODE'
;
;-----------------------------
;====== Program Segment Prefix
;-----------------------------
;
           ORG       002CH
ENVSEG     DW        ?                 ; Segment for environment
           ORG       00E0H
OLDINT09   LABEL     DWORD             ; Address of old INT 09H routine
  OLDI9_IP DW        ?                 ;   Offset
  OLDI9_CS DW        ?                 ;   Segment
INSCHAR    DB        ?                 ; Flag for inserting char into kb buffer
;
;----------------------------------
;====== Usual code for .COM program
;----------------------------------
;
           ORG       0100H
           ASSUME    CS:COMSEG, DS:COMSEG, ES:COMSEG
ENTER      LABEL     NEAR
           JMP       INIT
;
;====== Keyboard Interceptor
;
           ASSUME    CS:COMSEG, DS:NOTHING, ES:NOTHING
NEWINT09   LABEL     NEAR
;
;****** Initialization
;------ Enable interrupts, save flags and registers
           STI
           PUSHF
           PUSH      AX
           PUSH      DS
;------ Point DS at BIOS_DATA
           MOV       AX,BIOS_DATA      ; Point DS to BIOS data area
           MOV       DS,AX
           ASSUME    DS:BIOS_DATA
;------ Get scan code of keystroke
           IN        AL,KBDATA
;------ Assume char will not be put into keyboard buffer
           MOV       INSCHAR,0
;
;****** <Tilde> key pressed
TILDE      LABEL     NEAR
;------ Loop around if it's not <Tilde> key
           CMP       AL,29H
           JNZ       ESCAPE
;------ Branch if the <Alt> key is down
           TEST      KB_STATUS,08H
           JNZ       RESET_KB
;------ Set AX for keyboard buffer and branch
           MOV       INSCHAR,01H
           MOV       AX,011BH
           JMP       SHORT  RESET_KB
;
;****** <Escape> key pressed
ESCAPE     LABEL     NEAR
;------ Jump to BIOS if it's not <Escape> key
           CMP       AL,01H
           JNZ       JMPBIOS
;------ Branch if the <Ctrl> or <Alt> key is down
           TEST      KB_STATUS,0CH
           JNZ       RESET_KB
;------ Set AX for keyboard buffer and branch
           MOV       INSCHAR,01H
           TEST      KB_STATUS,03H
           JNZ       ESC1
           MOV       AX,2960H
           JMP       SHORT  RESET_KB
  ESC1:    MOV       AX,297EH
           JMP       SHORT  RESET_KB

;****** Exit to the original INT 9H interrupt
JMPBIOS    LABEL     NEAR
           POP       DS
           ASSUME    DS:NOTHING
           POP       AX
           POPF
           JMP       OLDINT09
;
;****** Reset the keyboard
           ASSUME    DS:BIOS_DATA
RESET_KB   LABEL     NEAR
           PUSH      AX
           IN        AL,KBCTRL         ; Get control port value
           MOV       AH,AL             ; Save it
           OR        AL,80H            ; Set high bit of control value
           OUT       KBCTRL,AL         ; Send reset value
           MOV       AL,AH             ; Get original control value
           OUT       KBCTRL,AL         ; Enable keyboard
           POP       AX
;
;****** Insert the character code in AX into the keyboard buffer.
INS_CHAR   LABEL     NEAR
;------ Branch if INSCHAR not set
           CMP       INSCHAR,00H
           JZ        END_INT
;------ Save registers
           PUSH      BX                ; Save DX and DS
           PUSH      DX                ; Save DX and DS
;------ See if there's room in the buffer
           MOV       BX,BUF_TAIL       ; Get current buffer tail address
           MOV       DX,BX             ; Transfer it to DX
           ADD       DX,2              ; Calculate next buffer position
           CMP       DX,BUF_END        ; Did we overshoot the end?
           JB        INSERT1           ;   No, then continue
           MOV       DX,BUF_START      ;   Yes, then wrap around
INSERT1:   CMP       DX,BUF_HEAD       ; Is the buffer full?
           JE        INSERT2           ;   Yes, then branch
;------ Place AX in buffer and advance buffer tail
           MOV       [BX],AX
           MOV       BUF_TAIL,DX
;------ Restore registers
INSERT2    LABEL     NEAR
           POP DX
           POP BX
;
;****** Enable interrupts and return to caller
END_INT    LABEL     NEAR
;------ Enable hardware interrupts
           MOV       AL,EOI            ; Issue EOI to 8259 controller
           OUT       INT_CTRL,AL
;------ Restore registers and return
           POP       DS
           ASSUME    DS:NOTHING
           POP       AX                ; Clean up the stack
           POPF
           IRET                        ; Return from interrupt
;
;====== Installation routine
;
           ASSUME    CS:COMSEG, DS:COMSEG, ES:COMSEG
INIT       LABEL     NEAR
;
;****** Revector INT 09
;------ Get address of existing keyboard interrupt routine
           MOV       AX,3509H
           INT       21H
           MOV       OLDI9_IP,BX
           MOV       OLDI9_CS,ES
;------ Put address of our routine IN the interrupt vectors
           MOV       DX,OFFSET NEWINT09
           MOV       AX,2509H
           INT       21H
;
;****** Normal TSR termination
;------ Free environment memory block
           MOV       AX,ENVSEG
           MOV       ES,AX
           ASSUME    ES:NOTHING
           MOV       AH,49H
           INT       21H
;------ Install as TSR
           MOV       DX,OFFSET INIT
           INT       27H

COMSEG     ENDS
           END      ENTER
